#!/usr/bin/perl -T

use strict;
use warnings;

#*******************************************************************************
# Mentat::Storage::Mongo class tests
#
# Author:   Jan Mach, jan.mach@cesnet.cz
# Version:  0.1
# Purpose:
# Comments:
#*******************************************************************************

use Data::Dumper;
#use Smart::Comments;

use Test::More tests => 89;

# First test if the module correctly compiles
BEGIN {
        use vars qw($TESTED_CLASS);
        $TESTED_CLASS = 'Mentat::Storage::Mongo';

        use_ok('Tester');
        use_ok('Mentat::Storage');
        use_ok('Value::Convertor');
        use_ok($TESTED_CLASS);
    }

# Debug output explicit ON switch
#Tester::DEBUG_ON();

# Debug output explicit OFF switch
#Tester::DEBUG_OFF();

# User should know, which version is beeing tested (just for sure)
diag("$TESTED_CLASS version: " . $Mentat::Storage::Mongo::VERSION) if Tester::D();

# Now attempt to create a class instance and check it
my $instance = $TESTED_CLASS->new(database => 'mentat', collection => 'unittest');
isa_ok($instance, $TESTED_CLASS);

my ($count, $message);
eval {

    # Count current number of records in database
    $count = $instance->count();
    diag("Database currently contains $count record(s)\n") if Tester::D();

    # Check storing to database
    my $record_id1 = $instance->insert({a => 1, b => 1});
    isnt($record_id1, undef, 'Insert test 01');

    my $record_id2 = $instance->insert({a => 2, b => 2});
    isnt($record_id2, undef, 'Insert test 02');

    my $record_id3 = $instance->insert({a => 3, b => 3});
    isnt($record_id3, undef, 'Insert test 03');

    my $record_id4 = $instance->insert({a => 4, b => 4});
    isnt($record_id4, undef, 'Insert test 04');

    my $record_id5 = $instance->insert({a => 5, b => 5});
    isnt($record_id5, undef, 'Insert test 04');

    my $record_id6 = $instance->insert({a => 6, b => 6});
    isnt($record_id6, undef, 'Insert test 04');

    my $record_id7 = $instance->insert({a => 7, b => 7});
    isnt($record_id7, undef, 'Insert test 04');

    my $record_id8 = $instance->insert({a => 8, b => 8});
    isnt($record_id8, undef, 'Insert test 04');

    is($instance->count(), $count + 8, 'Count test 01');

    # Check fetching from database
    is(ref $instance->fetch($record_id1), 'HASH', 'Fetch test 01');
    is(ref $instance->fetch($record_id2), 'HASH', 'Fetch test 02');
    is(ref $instance->fetch($record_id3), 'HASH', 'Fetch test 03');
    is(ref $instance->fetch($record_id4), 'HASH', 'Fetch test 04');
    is(ref $instance->fetch($record_id1), 'HASH', 'Fetch test 05');
    is(ref $instance->fetch($record_id2), 'HASH', 'Fetch test 06');
    is(ref $instance->fetch($record_id3), 'HASH', 'Fetch test 07');
    is(ref $instance->fetch($record_id4), 'HASH', 'Fetch test 08');
    is($instance->fetch('foobar'),        undef,  'Fetch test 09');

    # Check updating the records in database
    my $dr1 = $instance->update({a => {'$lte'=>2}},{'$inc' => {b => 3 }});
    is_deeply([$dr1->{'ok'}, $dr1->{'n'}, $dr1->{'err'}], ['1', 1, undef], "Update test 01");
    is_deeply([$instance->find({'_id' => $record_id1})],[[{_id => $record_id1, a => 1, b => 4}],1],'Update test 02');
    is_deeply([$instance->find({'_id' => $record_id2})],[[{_id => $record_id2, a => 2, b => 2}],1],'Update test 03');
    my $dr2 = $instance->update({a => {'$lte'=>2}},{'$inc' => {b => 3}},{multiple => 1});
    is_deeply([$dr2->{'ok'}, $dr2->{'n'}, $dr2->{'err'}], ['1', 2, undef], "Update test 04");
    is_deeply([$instance->find({'_id' => $record_id1})],[[{_id => $record_id1, a => 1, b => 7}],1],'Update test 05');
    is_deeply([$instance->find({'_id' => $record_id2})],[[{_id => $record_id2, a => 2, b => 5}],1],'Update test 06');
    my $dr3 = $instance->update({'_id' => $record_id1},{a => 'x', b => 'y'});
    is_deeply([$dr3->{'ok'}, $dr3->{'n'}, $dr3->{'err'}], ['1', 1, undef], "Update test 07");
    is_deeply([$instance->find({'_id' => $record_id1})],[[{_id => $record_id1, a => 'x', b => 'y'}],1],'Update test 08');

    # Check deleting from database
    my $dr4 = $instance->delete($record_id1);
    my $dr5 = $instance->delete($record_id2);
    my $dr6 = $instance->delete('bogus__id');
    is_deeply([$dr4->{'ok'}, $dr4->{'n'}, $dr4->{'err'}], ['1', 1, undef], "Delete test 01");
    is_deeply([$dr5->{'ok'}, $dr5->{'n'}, $dr5->{'err'}], ['1', 1, undef], "Delete test 02");
    is_deeply([$dr6->{'ok'}, $dr6->{'n'}, $dr6->{'err'}], ['1', 0, undef], "Delete test 03");

    # Check counting from database
    is($instance->count({'_id' => $record_id1}), 0, 'Count test 01');
    is($instance->get_last_filter(), '[COUNT] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id1.'"}\'', 'Last filter test 01');
    is($instance->count({'_id' => $record_id2}), 0, 'Count test 02');
    is($instance->get_last_filter(), '[COUNT] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id2.'"}\'', 'Last filter test 02');
    is($instance->count({'_id' => $record_id3}), 1, 'Count test 03');
    is($instance->get_last_filter(), '[COUNT] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id3.'"}\'', 'Last filter test 03');
    is($instance->count({'_id' => $record_id4}), 1, 'Count test 04');
    is($instance->get_last_filter(), '[COUNT] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id4.'"}\'', 'Last filter test 04');
    is($instance->count(), $count + 6, 'Count test 05');
    is($instance->get_last_filter(), '[COUNT] Database: \'mentat:unittest\', filter \'\'', 'Last filter test 05');

    # Expected results
    my $ER01 = [
            {_id=>"".$record_id3,a=>3,b=>3},
            {_id=>"".$record_id4,a=>4,b=>4},
            {_id=>"".$record_id5,a=>5,b=>5},
            {_id=>"".$record_id6,a=>6,b=>6},
            {_id=>"".$record_id7,a=>7,b=>7},
            {_id=>"".$record_id8,a=>8,b=>8},
        ];
    my $ER02 = [
            {_id=>"".$record_id5,a=>5,b=>5},
            {_id=>"".$record_id6,a=>6,b=>6},
            {_id=>"".$record_id7,a=>7,b=>7},
            {_id=>"".$record_id8,a=>8,b=>8},
        ];
    my $ER03 = [
            {_id=>"".$record_id3,a=>3,b=>3},
            {_id=>"".$record_id4,a=>4,b=>4},
        ];
    my $ER04 = [
            {_id=>"".$record_id8,a=>8,b=>8},
            {_id=>"".$record_id7,a=>7,b=>7},
            {_id=>"".$record_id6,a=>6,b=>6},
            {_id=>"".$record_id5,a=>5,b=>5},
        ];
    my $ER05 = [
            {_id=>"".$record_id4,a=>4,b=>4},
            {_id=>"".$record_id3,a=>3,b=>3},
        ];
    my $ER06 = [
            {_id=>"".$record_id8,a=>8,b=>8},
            {_id=>"".$record_id7,a=>7,b=>7},
        ];
    my $ER07 = [
            {_id=>"".$record_id4,a=>4,b=>4},
        ];
    my $ER08 = [
            {_id=>"".$record_id6,a=>6,b=>6},
            {_id=>"".$record_id5,a=>5,b=>5},
        ];
    my $ER09 = [
            {_id=>"".$record_id3,a=>3,b=>3},
        ];
    my $ER10 = [
            {_id=>"".$record_id7,a=>7,b=>7},
            {_id=>"".$record_id6,a=>6,b=>6},
        ];
    my $ER11 = [
            {_id=>"".$record_id3,a=>3,b=>3},
        ];
    my $ER12 = [
            {_id=>"".$record_id7,a=>7},
            {_id=>"".$record_id6,a=>6},
        ];
    my $ER13 = [
            {_id=>"".$record_id7,b=>7},
            {_id=>"".$record_id6,b=>6},
        ];

    my $result = [];
    sub fcb($) { push(@$result, @_); }
    my ($cursor, $cnta) = $instance->find_i(undef,undef,{'a' => 1});
    my $cntb = $instance->cursor_iterate($cursor, \&fcb);
    is_deeply([$result,$cnta],[$ER01,6],'Iterating find test 01');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'\' with sorting \'{"a":1}\'', 'Iterating find check 02');
    is($cnta, $cntb, "Iterating find check 02");

    # Check searching from database
    is_deeply([$instance->find(undef,undef,{'a' => 1})],[$ER01,6],'Find test 01');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'\' with sorting \'{"a":1}\' with limit \'500000\'', 'Last filter test 06');

    is_deeply([$instance->find({'_id' => $record_id1})], [[],0],'Find test 02');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id1.'"}\' with limit \'500000\'', 'Last filter test 07');

    is_deeply([$instance->find({'_id' => $record_id3})], [[{_id=>"".$record_id3,a=>3,b=>3}],1],'Find test 03');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id3.'"}\' with limit \'500000\'', 'Last filter test 08');

    is_deeply([$instance->find({'a' => 8})], [[{_id=>"".$record_id8,a=>8,b=>8}],1],'Find test 04');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":8}\' with limit \'500000\'', 'Last filter test 09');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},undef,{'a' => 1})], [$ER02,4],'Find test 05');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with sorting \'{"a":1}\' with limit \'500000\'', 'Last filter test 10');

    is_deeply([$instance->find({'a' => {'$lt' => 5}},undef,{'a' => 1})], [$ER03,2],'Find test 06');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$lt":5}}\' with sorting \'{"a":1}\' with limit \'500000\'', 'Last filter test 11');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},undef,{'b' => -1})], [$ER04,4],'Find test 07');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with sorting \'{"b":-1}\' with limit \'500000\'', 'Last filter test 12');

    is_deeply([$instance->find({'a' => {'$lt' => 5}},undef,{'b' => -1})], [$ER05,2],'Find test 08');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$lt":5}}\' with sorting \'{"b":-1}\' with limit \'500000\'', 'Last filter test 13');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},undef,{'b' => -1},2)], [$ER06,4],'Find test 09');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with sorting \'{"b":-1}\' with limit \'2\'', 'Last filter test 14');

    is_deeply([$instance->find({'a' => {'$lt' => 5}},undef,{'b' => -1},1)], [$ER07,2],'Find test 10');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$lt":5}}\' with sorting \'{"b":-1}\' with limit \'1\'', 'Last filter test 15');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},undef,{'b' => -1},undef,2)], [$ER08,4],'Find test 11');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with sorting \'{"b":-1}\' with limit \'500000\' with skip \'2\'', 'Last filter test 16');

    is_deeply([$instance->find({'a' => {'$lt' => 5}},undef,{'b' => -1},undef,1)], [$ER09,2],'Find test 12');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$lt":5}}\' with sorting \'{"b":-1}\' with limit \'500000\' with skip \'1\'', 'Last filter test 17');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},undef,{'b' => -1},2,1)], [$ER10,4],'Find test 13');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with sorting \'{"b":-1}\' with limit \'2\' with skip \'1\'', 'Last filter test 18');

    is_deeply([$instance->find({'a' => {'$lt' => 5}},undef,{'b' => -1},1,1)], [$ER11,2],'Find test 14');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$lt":5}}\' with sorting \'{"b":-1}\' with limit \'1\' with skip \'1\'', 'Last filter test 19');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},{'a' => 1},{'b' => -1},2,1)], [$ER12,4],'Find test 15');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with projection \'{"a":1}\' with sorting \'{"b":-1}\' with limit \'2\' with skip \'1\'', 'Last filter test 20');

    is_deeply([$instance->find({'a' => {'$gt' => 4}},{'a' => 0},{'b' => -1},2,1)], [$ER13,4],'Find test 16');
    is($instance->get_last_filter(), '[FIND] Database: \'mentat:unittest\', filter \'{"a":{"$gt":4}}\' with projection \'{"a":0}\' with sorting \'{"b":-1}\' with limit \'2\' with skip \'1\'', 'Last filter test 21');

    # Try removing item from database
    my $rr1 = $instance->remove({'_id' => $record_id3});
    is_deeply([$rr1->{'ok'}, $rr1->{'n'}, $rr1->{'err'}], ['1', 1, undef], "Remove test 01");
    is($instance->get_last_filter(), '[REMOVE] Database: \'mentat:unittest\', filter \'{"_id":"'.$record_id3.'"}\'', 'Last filter test 20');

    my $rr2 = $instance->remove({'a' => {'$gt' => 5}});
    is_deeply([$rr2->{'ok'}, $rr2->{'n'}, $rr2->{'err'}], ['1', 3, undef], "Remove test 02");
    is($instance->get_last_filter(), '[REMOVE] Database: \'mentat:unittest\', filter \'{"a":{"$gt":5}}\'', 'Last filter test 21');

    is($instance->count({'_id' => $record_id3}), 0,          'Count test 05');
    is($instance->count({'_id' => $record_id4}), 1,          'Count test 06');
    is($instance->count(),                       $count + 2, 'Count test 07');

    # Try removing all items from database
    my $rr3 = $instance->remove();
    is_deeply([$rr3->{'ok'}, $rr3->{'n'}, $rr3->{'err'}], ['1', $count + 2, undef], "Remove test 03");
    is($instance->get_last_filter(), '[REMOVE] Database: \'mentat:unittest\', filter \'\'', 'Last filter test 22');
    is($instance->count(), 0, 'Count test 08');

};
if($@) {
    print STDERR "MongoDB error: $@\n";
}
